import os
import sqlite3
import random
import string
import logging

DB_DIR = '/var/spool/kepi'
DB_FILENAME = 'kepi.sqlite3'

STATES = {'unverified', 'inbox', 'received',
           'outbox', 'sent', 'gone',
           'errors',
           }

TABLE_CREATION = [
        (
            'create table message (local_id, content, state)'
            ),
        'create table waiting_on (url, local_id)',
        ]

logger = logging.getLogger('kepi.daemon')

class Spool:
    def __init__(self,
                 path=DB_DIR,
                 ):

        def mkdir(path):
            try:
                os.makedirs(path)
            except FileExistsError:
                pass

        self.path = path

        self.db = sqlite3.connect(
                os.path.join(path, DB_FILENAME)
                )

        self._ensure_tables()

    def _db_exec(self, sql, args=(), cu=None, ignore_errors=False):
        if cu is None:
            cu = self.db.cursor()

        try:
            result = cu.execute(sql, args)
            logger.debug("> %s %s", sql, args)
        except sqlite3.OperationalError as e:
            if ignore_errors:
                pass
            else:
                logger.debug("> %s %s --> error: %s",
                             sql, args, e)
                raise

        return result

    def _ensure_tables(self):
        cu = self.db.cursor()

        for sql in TABLE_CREATION:
            self._db_exec(sql, cu=cu, ignore_errors=True)

    def _make_local_id(self):
        result = ''.join([
                random.choice(string.ascii_letters)
                for i in range(8)
                ])
        return result

    def _make_spooledfile_object(self,
                                 local_id,
                                 message,
                                 state,
                                ):
        result = SpooledFile.__new__(SpooledFile)
        result._spool = self
        result._local_id = local_id
        result._message = message
        return result

    def add_incoming(self,
                     message,
                     starting_state = 'unverified',
                     ):
        assert isinstance(message, str)
        assert starting_state in STATES

        local_id = self._make_local_id()

        self._db_exec(
                'insert into message values (?, ?, ?)',
                (
                    local_id,
                    message,
                    starting_state,
                    ),
                )
        return self._make_spooledfile_object(
                local_id = local_id,
                message = message,
                state = starting_state,
                )

    def items(self):
        for local_id, content, state in self._db_exec(
                'select local_id, content, state from message'
                ):
            yield (local_id,
                   self._make_spooledfile_object(
                       local_id = local_id,
                       message = content,
                       state = state,
                       ))

    def __repr__(self):
        return f'[spool;path={self.path}]'

class SpooledFile:
    def __init__(self):
        raise ValueError(
                "Don't instantiate this."
                )

    def __eq__(self, other):
        if isinstance(other, str):
            return self._local_id == other
        elif isinstance(other, SpooledFile):
            return self._local_id == other._local_id
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def spool(self):
        return self._spool

    @property
    def message(self):
        return self._message

    @property
    def state(self):
        found = self._spool._db_exec(
                'select state from message where local_id=?',
                (self._local_id,),
                ).fetchone()
        return found[0]

    @state.setter
    def state(self, v):
        if v not in STATES:
            raise ValueError(f"{v} is not a state (they are {STATES})")
        assert v in STATES
        self._spool._db_exec(
                'update message set state=? where local_id=?',
                (v, self._local_id,),
                )

    @property
    def local_id(self):
        return self._local_id

    def __repr__(self):
        return f'[spooled file {self.local_id}]'
